from abc import abstractmethod, ABC
from typing import Callable, Dict
from asyncio import AbstractEventLoop
import zmq
import zmq.asyncio
import asyncio
from backtesterrb30.libs.utils.list_of_services import SERVICES
from backtesterrb30.libs.utils.service import Service
from backtesterrb30.libs.interfaces.utils.config import Config
from os import _exit
from pydantic import BaseModel
from json import dumps, loads


class ZMQ(Service, ABC):

    __is_running: bool
    __is_active: bool

    __commands: Dict[str, Callable]

    # override
    def __init__(self, config: Config, logger=print):
        super().__init__(config, logger)
        self.config=config
        self.__is_running = False
        self.__is_active = False

        self.__pub = None
        self.__subs = []
        

        self.__commands = {'start': self._start,
                          'pause': self._pause,
                          'stop': self._stop}

    # override
    def run(self):
        self.__init_sockets(self.config)
        self.__is_running = True
        self.__is_active = True
        super().run()


    # override
    def _send(self, service: SERVICES, msg: str, *args):
        if self.__is_active:
            data = [service.value.encode('utf-8'), msg.encode('utf-8')]
            for arg in args:
                if isinstance(arg,BaseModel):
                    arg = arg.dict()
                arg = dumps(arg, default=str)
                data.append(arg.encode('utf-8'))
            self.__pub.send_multipart(data)


    @abstractmethod
    def _handle_zmq_message(self, msg: str):
        pass

    @abstractmethod
    def _asyncio_loop(self, loop:asyncio.AbstractEventLoop):
        pass
    
    def _loop(self):
        loop = asyncio.get_event_loop()
        self._asyncio_loop(loop)
        loop.run_forever()
        loop.close()

    def _register(self, command: str, func: Callable):
        self.__commands[command] = func


    def _create_listeners(self, loop:AbstractEventLoop):
        self._log('Start main loop')
        for sub in self.__subs:
            loop.create_task(self.__listen_zmq(sub))


    async def __listen_zmq(self, sub):
        while self.__is_running:

            if self.__is_active:
                data = await sub.recv_multipart()
                if data:
                    await self.__handle(data)
                await asyncio.sleep(0.00001)
            else:
                await asyncio.sleep(0.01)
        self.__deinit()


    def __init_sockets(self, config: Config):
        sub_context = zmq.asyncio.Context()
        pub_context = zmq.Context()
        for sub in config.sub:
            s = sub_socket(sub_context, 'tcp://' + config.ip + ':' + str(sub.port), sub.topic)
            self.__subs.append(s)
        self.__pub = pub_socket(pub_context, 'tcp://*:' + str(config.pub.port))


    async def __handle(self, data):
        data = [x.decode('utf-8') for x in data]
        if len(data) >= 2:
            topic, cmd, *args = data
            func = self.__commands.get(cmd)
            if func:
                # self._log(f'Receive "{cmd}" command')
                if args:
                    args_loaded = []
                    for arg in args:
                        args_loaded.append(loads(arg))
                    await func(*args_loaded)
                else:
                    await func()
            else:
                self._log(f"Command '{cmd}' not registered")
        self._handle_zmq_message(data)


    def __deinit(self):
        self.__pub.close()


    def _stop(self):
        self._log("Service stoped")
        self.__is_running = False
        _exit(1)


    def _start(self):
        self._log("Service started")
        self.__is_active = True


    def _pause(self):
        self._log("Service paused")
        self.__is_active = False


def sub_socket(context: zmq.asyncio.Context(), url: str, topic: str):
    sub = context.socket(zmq.SUB)
    sub.connect(url)
    sub.subscribe(topic)
    return sub

def pub_socket(context: zmq.Context(), url: str):
    pub = context.socket(zmq.PUB)
    ret = pub.bind(url)
    return pub





import asyncio
from typing import List
from backtesterrb30.historical_data_feeds.historical_downloader import HistoricalDownloader
from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.data_sources.data_sources_list import HISTORICAL_SOURCES
from backtesterrb30.libs.interfaces.python_backtester.data_start import DataStart
from backtesterrb30.libs.communication_broker.zmq_broker import ZMQ
from backtesterrb30.libs.communication_broker.asyncio_broker import AsyncioBroker
from backtesterrb30.libs.interfaces.utils.config import Config, BROKERS
from backtesterrb30.libs.utils.list_of_services import SERVICES, SERVICES_ARRAY
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import InstrumentFile
from backtesterrb30.historical_data_feeds.functions import get_instrument_files, synchronize_dataframes, load_data_frame_ticks_2
from datetime import datetime
from os import path, makedirs
from os import listdir
from os.path import isfile, join
import pandas as pd
import time as tm
from backtesterrb30.libs.utils.service import Service
from appdirs import user_cache_dir
from json import dumps, loads

class HistoricalDataFeeds(Service):
    
    downloaded_data_path = user_cache_dir('rb30_cache')
    _broker: BrokerBase

    def __init__(self, config: Config, data_schema: DataSchema, loop = None, logger=print):
        super().__init__(config, logger)
        self.config: Config=config
        self.__loop =  loop
        self.__custom_event_loop = False
        if self.__loop == None: 
            self.__loop = asyncio.get_event_loop()
            self.__custom_event_loop= True
        self.__data_schema: DataSchema = data_schema
        self.__columns=['timestamp']+[c.symbol for c in self.__data_schema.data]
        self.__historical_sources_array = [i for i in dir(HISTORICAL_SOURCES) if not i.startswith('__')]
        self.__data_sources = {}
        self.__sending_locked = False
        self.__start_time = 0
        self.__data_to_download_2 = None
        self.__engine_ready = False
        self.__historical_downloader = HistoricalDownloader(self.__data_schema, 
                self.__historical_sources_array, self.downloaded_data_path, self._log)



    # # override
    # def _asyncio_loop(self, loop: asyncio.AbstractEventLoop):
    #     self._create_listeners(loop)
    #     self.__validate_downloaded_data_folder()
    #     if self.config.backtest:
    #         self.__data_to_download_2 = self.__historical_downloader.run(loop)
    #         data_parts = self.__prepare_loading_data_structure_2()
    #         loop.create_task(self.__historical_data_loop_ticks(data_parts))
    #     else:  
    #         self._log('backtest is off')

    def _loop(self):
        # self._broker.run()
        self._broker.create_listeners(self.__loop)
        self.__validate_downloaded_data_folder()
        if self.config.backtest:
            self.__data_to_download_2 = self.__historical_downloader.run(self.__loop)
            data_parts = self.__prepare_loading_data_structure_2()
            self.__loop.create_task(self.__historical_data_loop_ticks(data_parts))
        else:  
            self._log('backtest is off')
        if self.__custom_event_loop:
            self.__loop.run_forever()
            self.__loop.close()

    # def _send(self, service: SERVICES, msg: str, *args):
    #     self._broker.send(service, msg, *args)

    def _configure(self):
        super()._configure()
        # register commands
        self._broker.register("unlock_historical_sending", self.__unlock_historical_sending_event)
        self._broker.register("engine_ready_response", self.__engine_ready_response_event)

    # # override
    # def _handle_zmq_message(self, message):
    #     pass


    async def __send_start_params(self):
        self.__start_time = tm.time()
        file_names_grouped = []
        for symbol in self.__data_schema.data:
            file_names_grouped.append({
                "symbol": symbol.symbol,
                "source": symbol.historical_data_source,
                "files": get_instrument_files(symbol)
            })
        start_params = {
            'file_names': file_names_grouped,
            'start_time': self.__start_time
        }
        await self._broker.send(SERVICES.python_backtester, 'data_start', DataStart(**start_params))
        return


    async def __historical_data_loop_ticks(self, data_parts: dict):
        # waiting for zero mq ports starts up
        await asyncio.sleep(0.5)
        while True:
            # if self.__data_downloaded(self.__data_to_download):
            await self._broker.send(SERVICES.python_engine, 'engine_ready', SERVICES.historical_data_feeds.value)
            if self.__validate_data_downloaded(self.__data_to_download_2) and self.__engine_ready: 
                self._log('All data has been downloaded')
                
                sending_counter = 0
                self._log('Starting data loop')
                await self.__send_start_params()
                # return
                last_row = []
                for _, one_year_array in data_parts.items():
                    self._log('Synchronizing part of data')
                    data_part = load_data_frame_ticks_2(self.__data_schema, self.__columns, self.downloaded_data_path, last_row, one_year_array)
                    for row in data_part:
                        last_row = row
                        await self._broker.send(SERVICES.python_engine,'data_feed',list(last_row))
                        sending_counter += 1
                        if sending_counter % 1000 == 0:
                            self.__sending_locked = True
                            await self._broker.send(SERVICES.python_engine, 'historical_sending_locked')
                            while self.__sending_locked:
                                await asyncio.sleep(0.01)

                await self._broker.send(SERVICES.python_engine, 'data_finish')
                break
            await asyncio.sleep(1)


    def __validate_data_downloaded(self, files_to_download: List[InstrumentFile]):
        if files_to_download == None:
            return False
        files_in_directory = [f for f in listdir(self.downloaded_data_path) if isfile(join(self.downloaded_data_path, f))]
        data_to_download = list(set([f.to_filename() for f in files_to_download]) - set(files_in_directory))
        return data_to_download == []


    def __validate_downloaded_data_folder(self):
        if not path.exists(self.downloaded_data_path):
            makedirs(self.downloaded_data_path)


    def __prepare_loading_data_structure_2(self) -> dict:
        files_collection = {}
        for symbol in self.__data_schema.data:
            files: List[InstrumentFile] = get_instrument_files(symbol)
            for f in files:
                if not f.time_stop in files_collection:
                    files_collection[f.time_stop]: List[InstrumentFile] = []
                files_collection[f.time_stop].append(f)
        return dict(sorted(files_collection.items()))

    async def __stop_all_services(self):
        for service in SERVICES_ARRAY:
            if service != self.name:
                await self._broker.send(getattr(SERVICES, service), 'stop')
        self._broker.stop()

    # COMMANDS
    
    async def __unlock_historical_sending_event(self):
        self.__sending_locked = False

    async def __engine_ready_response_event(self):
        self.__engine_ready = True



    # def __map_raw_to_instruments(self, raw: list, instruments: list):
    #     last_raw_obj = {}
    #     if len(raw) != len(instruments):
    #         raise Exception('Error in map_raw_to_instruments. Lengths of raw and list of instruments are not equal')
    #     for value, instrument in zip(raw, instruments):
    #         last_raw_obj[instrument] = value
    #     return last_raw_obj


    # def __prepare_dataframes_to_synchronize_2(self, downloaded_data_path: str, 
    #                 last_row: list, files_array: List[InstrumentFile]) -> List[dict]:
    #     list_of_dfs = []
    #     for data_element in self.__data_schema.data:
    #         columns = ['timestamp', data_element.symbol]
    #         file_name = 'none'
    #         actual_raw = [0,0]
    #         prev_raw = [0,0]
    #         for element in files_array:
    #             if data_element.symbol == element.instrument:
    #                 file_name = element.to_filename()
    #         if file_name == 'none':
    #             # No file in this period for this instrument. Set empty dataframe.
    #             df = pd.DataFrame([], columns=columns)
    #         else:
    #             # File exists. Load dataframe.
    #             df = pd.read_csv(join(downloaded_data_path, file_name), index_col=None, header=None, names=columns)
    #             # append last raw it if exists
    #             if last_row != []:
    #                 # self._log('appending last row')
    #                 last_raw_mapped = self.__map_raw_to_instruments(last_row, self.__columns)
    #                 prev_raw[0] = last_raw_mapped["timestamp"]
    #                 prev_raw[1] = last_raw_mapped[data_element.symbol]
    #         obj = {
    #             "trigger_feed": data_element.trigger_feed,
    #             "rows_iterator": df.iterrows(),
    #             "actual_raw": actual_raw,
    #             "prev_raw": prev_raw,
    #             "consumed": False
    #         }            
    #         #prepare to load:
    #         if obj['actual_raw'][0] == 0:
    #             try:
    #                 i, v = next(obj['rows_iterator'])
    #                 obj['actual_raw'] = list(v)
    #             except StopIteration:
    #                 obj['consumed'] = True
    #         list_of_dfs.append(obj)
    #     return list_of_dfs
                

    # def __load_data_frame_ticks_2(self, downloaded_data_path: str, last_row: list, files_array: List[InstrumentFile]) -> List[list]:
    #     """
    #     Function is geting files array from one period that are going to be loaded. 
    #     Function returns synchronized data in list of lists which are ready to send to engine.
    #     """
    #     list_of_dfs = self.__prepare_dataframes_to_synchronize_2(downloaded_data_path, last_row, files_array)
    #     rows = synchronize_dataframes(list_of_dfs, last_row)
    #     return rows



    
    # def __get_data_to_download(self, data_schema: DataSchema):
    #     for data_symbol in data_schema.data:
    #         data_symbol.additional_properties['files_to_download']: List[InstrumentFile] = []
    #         data_symbol.additional_properties['files_to_download'] = self.__check_symbol_data_exists(data_symbol)
    #         if self.__data_to_download_2 == None:
    #             self.__data_to_download_2 = []
    #         self.__data_to_download_2 = self.__data_to_download_2 + data_symbol.additional_properties['files_to_download']


    # def __create_downloading_clients(self, historical_sources_array: list, data_schema: DataSchema, data_sources: dict):
    #     for source in historical_sources_array:
    #         if source in [data.historical_data_source for data in data_schema.data if \
    #                 data.additional_properties['files_to_download'] != []]:
    #             data_sources[source]: DataSource = getattr(HISTORICAL_SOURCES, source)(self._log)


    # def __validate_symbols_to_download(self,data_schema: DataSchema, loop: asyncio.AbstractEventLoop):
    #     self._log('validating symbols before download')
    #     for symbol in data_schema.data:
    #         if symbol.additional_properties['files_to_download'] != []:
    #             data_source_client: DataSource = self.__get_data_source_client(symbol.historical_data_source)
    #             res = loop.run_until_complete(data_source_client.validate_instrument(symbol))
    #             if not res:
    #                 raise Exception('Error while validation symbol.')

    # def __download_prepared_data(self, data_schema: DataSchema, loop: asyncio.AbstractEventLoop):
    #     for symbol in data_schema.data:
    #         if symbol.additional_properties['files_to_download'] != []:
    #             loop.create_task(self.__download_symbol_data(symbol))


    # def __get_data_source_client(self, historical_source: str) -> DataSource:
    #     client = self.__data_sources[historical_source]
    #     if not client:
    #             self._log('Error, no registered source client')
    #             self.__stop_all_services()
    #     return client


    # def __check_symbol_data_exists(self, data_symbol: DataSymbol) -> List[InstrumentFile]:
    #     """
    #     data scheme
    #     <symbol>__<source>__<interval>__<date-from>__<date-to>
    #     all instruments are downloaded in year files.
    #     """
    #     files: List[InstrumentFile] = get_instrument_files(data_symbol)
    #     # self._log('files to download "'+str(data_symbol.symbol)+'" :', files)
    #     files_in_directory = [f for f in listdir(self.downloaded_data_path) if isfile(join(self.downloaded_data_path, f))]
    #     files_to_download = list(set([f.to_filename() for f in files]) - set(files_in_directory))
    #     files_to_download = [InstrumentFile.from_filename(file) for file in files_to_download]
    #     return files_to_download
     

    # async def __download_symbol_data(self, symbol: DataSymbol):
    #     files_to_download: List[InstrumentFile] = symbol.additional_properties['files_to_download']
    #     await self.__download_symbol_files(files_to_download, symbol.historical_data_source)


    # async def __download_symbol_files(self, files: List[InstrumentFile], source: str):
    #     data_source_client: DataSource = self.__get_data_source_client(source)
    #     for file in files:
    #         await data_source_client.download_instrument(self.downloaded_data_path, file)


    # def __validate_data_schema_instruments(self, data_symbol_array: List[DataSymbol], loop: asyncio.AbstractEventLoop):
    #     self._log('Data_schema validation')
    #     #check for duplicates:
    #     seen = []
    #     for x in data_symbol_array:
    #         data = str(x.symbol)+str(x.historical_data_source)
    #         if data in seen:
    #             raise Exception('Duplicate data symbols')
    #         else:
    #             seen.append(data)
    #     #check other
    #     number_of_trigger_feeders = 0
    #     for data in data_symbol_array:
    #         if data.backtest_date_start == None:
    #             raise Exception('Error. You must provide "backtest_date_start" field in data_schema file while you are backtesting your strategy')
    #         if data.backtest_date_start >= data.backtest_date_stop: 
    #             raise Exception('Error. You have provided "backtest_date_start" is equal or bigger than "backtest_date_start" ')
    #         if [data.backtest_date_start.hour,
    #             data.backtest_date_start.minute,
    #             data.backtest_date_start.second,
    #             data.backtest_date_start.microsecond] != [0,0,0,0]:
    #             raise Exception('Error. Provide your "backtest_date_start" and "backtest_date_stop" in a day accuracy like: "backtest_date_start": datetime(2020,6,1)')
    #         if data.historical_data_source not in self.__historical_sources_array: 
    #             raise Exception('Error. This historical_data_source not implemented yet')
    #         if data.trigger_feed == True:
    #             number_of_trigger_feeders += 1
    #     if number_of_trigger_feeders < 1:
    #         raise Exception('Error. Your "data_schema.py" must have at least one instrument that triggers feeds')

    # # override
    # def _loop(self):
    #     loop = asyncio.get_event_loop()
    #     self._create_listeners(loop)
    #     self.__validate_downloaded_data_folder()
    #     if self.config.backtest:
    #         self.__get_data_to_download(self.__data_schema)
    #         self.__create_downloading_clients(self.__historical_sources_array, self.__data_schema, self.__data_sources)
    #         self.__validate_data_schema_instruments(self.__data_schema.data, loop)
    #         self.__validate_symbols_to_download(self.__data_schema, loop)
    #         self.__download_prepared_data(self.__data_schema, loop)

    #         data_parts = self.__prepare_loading_data_structure_2()
    #         loop.create_task(self.__historical_data_loop_ticks(data_parts))
    #     else:  
    #         self._log('backtest is off')
    #     loop.run_forever()
    #     loop.close()



from abc import abstractmethod
import asyncio
from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import InstrumentFile
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.communication_broker.zmq_broker import ZMQ
from backtesterrb30.libs.data_sources.data_sources_list import HISTORICAL_SOURCES
from backtesterrb30.libs.utils.list_of_services import SERVICES
from backtesterrb30.historical_data_feeds.functions import load_data_frame_with_dfs
from backtesterrb30.libs.utils.service import Service
from backtesterrb30.libs.utils.timestamps import timestamp_to_datetime
from backtesterrb30.libs.interfaces.utils.config import Config, BROKERS
from os import getenv
from datetime import datetime
import time
from appdirs import user_cache_dir

class LiveDataFeeds(Service):
    """Python Live data provider"""

    downloaded_data_path = user_cache_dir('rb30_cache')
    _broker: BrokerBase
    
    def __init__(self, config: dict, data_schema: DataSchema, loop = None, logger=print):
        super().__init__(config, logger)
        self.config: Config=config
        self.__loop =  loop
        self.__custom_event_loop = False
        if self.__loop == None: 
            self.__loop = asyncio.get_event_loop()
            self.__custom_event_loop= True
        self.__data_schema: DataSchema = data_schema
        self.__columns=['timestamp']+[c.symbol for c in self.__data_schema.data]
        self.__historical_sources_array = [i for i in dir(HISTORICAL_SOURCES) if not i.startswith('__')]
        self.__data_sources = {}
        self.__download_buffer = True
        self.__engine_ready = False
        self.__historical_rows = []

    # override
    def _loop(self):
        self._broker.create_listeners(self.__loop)
        # TODO add data downloader
        if not self.config.backtest:
            self.__create_downloading_clients(self.__historical_sources_array, self.__data_schema, self.__data_sources, self._log)
            self.__combine_clients_and_symbols(self.__data_schema)
            interval = self.__find_loop_interval(self.__data_schema)
            self.__loop.create_task(self.__main_loop(interval))
        if self.__custom_event_loop:
            self.__loop.run_forever()
            self.__loop.close()

    def _configure(self):
        super()._configure()
        self._broker.register("engine_set_buffer_length", self.__engine_set_buffer_length_event)

    # def _send(self, service: SERVICES, msg: str, *args):
    #     self._broker.send(service, msg, *args)

    # # override
    # def _handle_zmq_message(self, message):
    #     pass 

    async def __main_loop(self, interval):
        if self.__download_buffer == True:
            await self._broker.send(SERVICES.python_engine, 'get_buffer_length', SERVICES.live_data_feeds.value)
        else:
            self.__engine_ready = True
        while True:
            if self.__engine_ready:
                timestamp = await self.__get_proper_timestamp(self.__historical_rows, self.__data_schema)
                for row in self.__historical_rows:
                    last_row = row
                    await self._broker.send(SERVICES.python_engine,'data_feed',list(last_row))
                while True:
                    row = []
                    row.append(timestamp)
                    sth_updated = False
                    for data_symbol in self.__data_schema.data:
                        price, updated = self.__get_symbol_price(timestamp, data_symbol)
                        if updated: sth_updated = True
                        row.append(price)
                    if sth_updated: 
                        await self._broker.send(SERVICES.python_engine,'data_feed',row)
                    await asyncio.sleep(interval/1000)
                    timestamp += interval
            await asyncio.sleep(2)

    def __get_smallest_interval(self, data_schema: DataSchema):
        smallest = 9999999999999
        for data_symbol in data_schema.data:
            client: DataSource = data_symbol.additional_properties['downloading_client']
            milis = client._get_interval_miliseconds(data_symbol.interval.value)
            if not milis or not (1000*60 <= milis <= 1000*60*60*24*6 ):
                raise Exception(data_symbol.symbol, 'Cant run live with this interval')
            if milis < smallest:
                smallest = milis
        return smallest

    async def __download_data(self, buffer_length: int, data_schema: DataSchema):
        buffer_length = buffer_length
        smallest = self.__get_smallest_interval(data_schema)
        time_stop = int(time.time()) * 1000 + smallest
        time_start = time_stop - 4 * smallest * buffer_length
        if time_stop - time_start < 1000*60*60*24: 
            time_start = time_stop - 1000*60*60*24
        dfs = []
        for data_symbol in data_schema.data:
            client: DataSource = data_symbol.additional_properties['downloading_client']
            df = await client.download_dataframe(
                    data_symbol.symbol,
                    data_symbol.interval.value,
                    time_start,
                    None)
            df.columns = ['timestamp', data_symbol.identifier]
            dfs.append(df)
        self.__historical_rows = load_data_frame_with_dfs(self.__data_schema, self.__columns, [], dfs)
        self.__historical_rows= self.__historical_rows[-buffer_length-1:]

    async def __get_proper_timestamp(self, historical_rows, data_schema: DataSchema):
        if historical_rows == []:
            return time.time() * 1000
        smallest = self.__get_smallest_interval(data_schema)
        current_time_with_reserve = time.time()*1000 + 1000
        next_should_be_timestamp = historical_rows[-1][0] + smallest
        if next_should_be_timestamp < current_time_with_reserve: 
            raise Exception('Did not managed to download historical buffer this time. Trying again')
        await asyncio.sleep(abs(next_should_be_timestamp - time.time()* 1000)/1000)
        return next_should_be_timestamp

    def __find_loop_interval(self, data_schema: DataSchema) -> int:
        smallest = self.__get_smallest_interval(data_schema)
        for data_symbol in data_schema.data:
            client: DataSource = data_symbol.additional_properties['downloading_client']
            milis = client._get_interval_miliseconds(data_symbol.interval.value)
            num = (milis / smallest ) % 1000
            if int(num) != num:
                raise Exception('Bad intervals... All intervals should be divisible by smallest one.')
        return smallest

    def __get_symbol_price(self, timestamp: int, symbol: DataSymbol):
        updated = False
        price = 0
        client: DataSource = symbol.additional_properties['downloading_client']
        if not 'last_price_timestamp' in symbol.additional_properties:
            symbol.additional_properties['last_price_timestamp'] = timestamp
            updated = True
        if not 'last_price' in symbol.additional_properties:
            symbol.additional_properties['last_price'] = client.get_current_price(symbol)
            updated = True
        milis = client._get_interval_miliseconds(symbol.interval.value)
        if timestamp - symbol.additional_properties['last_price_timestamp'] < milis:
            price = symbol.additional_properties['last_price']
        if timestamp - symbol.additional_properties['last_price_timestamp'] == milis:
            price = client.get_current_price(symbol)
            symbol.additional_properties['last_price_timestamp'] = timestamp
            symbol.additional_properties['last_price'] = price
            updated = True
        if timestamp - symbol.additional_properties['last_price_timestamp'] > milis:
            raise Exception('Error. Bad timestamp Such situation should not happen')
        return price, updated


    def __create_downloading_clients(self, historical_sources_array: list, data_schema: DataSchema, data_sources: dict, log = print):
        for source in historical_sources_array:
            if source in [data.historical_data_source for data in data_schema.data]:
                data_sources[source]: DataSource = getattr(HISTORICAL_SOURCES, source)(log)
            

    def __combine_clients_and_symbols(self, data_schema: DataSchema):
        for data_symbol in data_schema.data:
            data_symbol.additional_properties['downloading_client'] = self.__get_data_source_client(data_symbol.historical_data_source)

    def __get_data_source_client(self, historical_source: str) -> DataSource:
        client = self.__data_sources[historical_source]
        if not client:
                raise Exception('Error, no registered source client')
        return client
        
    #COMMANDS

    async def __engine_set_buffer_length_event(self, buffer_length: int):
        self.__buffer_length = buffer_length
        await self.__download_data(buffer_length, self.__data_schema)
        self.__engine_ready = True


from backtesterrb30.libs.utils.run_service import run_service
microservice_name = 'live_data_feeds'
from backtesterrb30.live_data_feeds.live_data_feeds import LiveDataFeeds
run_service(microservice_name, LiveDataFeeds)


import os
from backtesterrb30.libs.utils.module_loaders import import_executor_module
here = os.getcwd()
strategy_path = os.getenv('STRATEGY_PATH')
strategy_file = os.getenv('STRATEGY_FILE')
executor_class_name = os.getenv('EXECUTOR_CLASS_NAME')
from backtesterrb30.libs.utils.run_service import run_service
microservice_name = 'python_executor'
module = import_executor_module(strategy_path, strategy_file, executor_class_name)
run_service(microservice_name, module)



from abc import abstractmethod
import asyncio
import json
from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.interfaces.python_backtester.debug_breakpoint import DebugBreakpoint
from backtesterrb30.libs.interfaces.python_backtester.last_feed import LastFeed
from backtesterrb30.libs.interfaces.python_backtester.trade import Trade
from backtesterrb30.libs.interfaces.python_executor.executor_position import ExecutorPosition
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.communication_broker.zmq_broker import ZMQ
from backtesterrb30.libs.communication_broker.asyncio_broker import AsyncioBroker
from typing import List
from backtesterrb30.libs.utils.service import Service
from backtesterrb30.libs.utils.list_of_services import SERVICES
from backtesterrb30.libs.interfaces.utils.config import BROKERS, Config 

class Executor(Service):
    """Python Trade executor"""
    _broker: BrokerBase
    
    def __init__(self, config: Config, data_schema: DataSchema, loop = None, logger=print):
        super().__init__(config, logger)
        self.config: Config=config
        self.__loop =  loop
        self.__custom_event_loop = False
        if self.__loop == None: 
            self.__loop = asyncio.get_event_loop()
            self.__custom_event_loop= True
        self.__data_schema: DataSchema = data_schema
        self.__columns=['timestamp']+[c.symbol for c in self.__data_schema.data]
        # self.__event_price = 0
        # self.__event_timestamp = 0
        # self.__event_last_feed = []
        # self.__number_of_actions = 0
        self.__current_capital = 0
        self.__current_invested = 0
        self.__start_amout_of_usd = 10000
        
        self.__positions: List[ExecutorPosition] = [{
            'instrument': 'inst',
            'number_of_actions': 123,
            'current_value': 12
        }]


    # public methods:
    # ==================================================================

    @abstractmethod
    async def on_event(self, message):
        """Function is being called when :class:`Model` class triggers 
        `trigger_event()` function. 

        :param message: message defined in `model.py` in strategy files.
        :type message: Any
        """
        self._log('method should be implemented in strategy function')
        pass

    def get_data_schema(self):
        """Returns data_schema defined in your strategy files in `data_schema.py`
        Data schema contains list of :class:`DataSymbol` objects. In every :class:`DataSymbol` object, 
        you can use `get_buffer()` function to get current buffer connected to this symbol.

        :return: :class:`DataSchema` object combined with this strategy.
        :rtype: DataSchema
        """
        return self.__data_schema

    def get_data_symbol_by_custom_name(self, custom_name: str):
        """Returns a :class:`DataSymbol` object using `custom_name` defined
        in `data_schema.py`, `custom_name` must be unique.

        :param custom_name: custom name defined in `data_schema.py` file.
        :type custom_name: str
        :return: :class:`DataSymbol` object combined with privided custom_name
        :rtype: DataSymbol
        """
        if type(custom_name) != str:
            raise Exception('Provided name is not string')
        arr = [d for d in self.__data_schema.data if d.custom_name == custom_name]
        if len(arr) == 0:
            raise Exception('No data symbol with such custom name')
        if len(arr) >1 : 
            raise Exception('Two elements with the same custom name')
        return arr[0]


    async def trade(self, trade_value: float, data_symbol: DataSymbol, price: float = None, timestamp: int = None):
        """Triggers trade.

        :param trade_value: value of trade in dollars.
        :type trade_value: float
        :param data_symbol: :class:`DataSymbol` on which trade is going to be made.
        :type data_symbol: DataSymbol
        :param price: price of trade. Default is current price. Works only in backtest mode.
        :type price: float, optional
        :param timestamp: timestamp of trade. Default is current timestamp. Works only in backtest mode.
        :type timestamp: int, optional
        """
        if not price and not timestamp:
            price = 0 
            timestamp = 0
        if price and not timestamp or timestamp and not price:
            raise Exception('Provide both price and timestamp or none of them.')
        if trade_value > self.__start_amout_of_usd + self.__current_capital - self.__current_invested:
            raise Exception('To big amout of trade')
        if self.config.backtest:
            trade_params = {
                'value': trade_value,
                'price': price,
                'timestamp': timestamp,
                'symbol': data_symbol.symbol,
                'source': data_symbol.historical_data_source
            }
            await self._broker.send(SERVICES.python_backtester, 'trade', Trade(**trade_params))
            return True
        else:
            self._log('Live trading not implemented')
            pass


    async def close_all_trades(self):
        """Closes all opened trades.
        """
        if self.config.backtest == True:
            await self._broker.send(SERVICES.python_backtester, 'close_all_trades')
        else:
            # TODO trade in real broker
            pass


    # def _get_number_of_actions(self):
    #     return self.__number_of_actions

    # ==================================================================
    # end of public methods
    

    #private methods:

    # def _send(): pass
    # def _register(): pass
    # def _create_listeners(): pass

    def _loop(self):
        # self._broker.run()
        self._broker.create_listeners(self.__loop)
        if self.__custom_event_loop:
            self.__loop.run_forever()
            self.__loop.close()

    # def _send(self, service: SERVICES, msg: str, *args):
    #     self._broker.send(service, msg, *args)

    def _configure(self):
        super()._configure()
        self._broker.register("event", self.__event_event)
        # super()._register("set_number_of_actions", self.__set_number_of_actions_event)
        self._broker.register("set_current_capital_event", self.__set_current_capital_event)
        self._broker.register("set_current_invested_event", self.__set_current_invested_event)
        self._broker.register("debug_breakpoint", self.__debug_breakpoint_event)
        self._broker.register("last_feed", self.__last_feed_event)

    # # override
    # def _asyncio_loop(self, loop: asyncio.AbstractEventLoop):
    #     self._broker.create_listeners(loop)

    # # override
    # def _handle_zmq_message(self, message):
    #     pass      


    #COMMANDS

    async def __event_event(self, msg):
        await self.on_event(msg)

    # async def __set_number_of_actions_event(self, number: int):
    #     self.__number_of_actions = number

    async def __set_current_capital_event(self, number: int):
        self.__current_capital = number
    
    async def __set_current_invested_event(self, number: int):
        self.__current_invested = number

    async def __debug_breakpoint_event(self, breakpoint_params):
        await self._broker.send(SERVICES.python_backtester, 'debug_breakpoint', DebugBreakpoint(**breakpoint_params))

    async def __last_feed_event(self, msg):
        await self._broker.send(SERVICES.python_backtester, 'last_feed',LastFeed(**msg))


import os
from backtesterrb30.libs.utils.config_validator import validate_config
from backtesterrb30.libs.utils.module_loaders import import_data_schema
from backtesterrb30.libs.utils.service import Service
from backtesterrb30.libs.communication_broker.zmq_broker import ZMQ

def run_service(microservice_name: str, service_class: Service):
    here = os.getcwd()
    strategy_path = os.getenv('STRATEGY_PATH')
    sub_ports = [int(p) for p in os.getenv(microservice_name+'_subs').split(',')]
    pub_port = int(os.getenv(microservice_name+'_pubs'))
    strategy_file = os.getenv('STRATEGY_FILE')
    data_class_name = os.getenv('DATA_CLASS_NAME')
    backtest_state = os.getenv('backtest_state')
    from backtesterrb30.libs.interfaces.utils.config import Config

    config = {
        "name": microservice_name,
        "ip": "localhost",
        "sub": [
            {
            "topic": microservice_name,
            "port": p
            } for p in sub_ports
        ],
        "pub": {
                "topic": 'pub_'+str(pub_port),
                "port": pub_port
            } ,
        "backtest": True if backtest_state == 'true' else False,
        "debug": True,
        "strategy_path": strategy_path
    }
    data_class = import_data_schema(strategy_path, strategy_file, data_class_name)
    data_schema = validate_config(data_class.data)
    config = Config(**config)
    service: Service = service_class(config, data_schema)
    logger = service.get_logger()
    service.register_communication_broker(ZMQ(config, logger))
    service.run()


from pydantic import BaseModel
from typing import List
from backtesterrb30.libs.interfaces.utils.net_info import NetInfo
from typing import Optional
from enum import Enum

class BROKERS(Enum):
    zmq = 'zmq', 
    asyncio = 'asyncio'

class Config(BaseModel):
    name: str
    ip: Optional[str]
    sub: Optional[List[NetInfo]]
    pub: Optional[NetInfo]
    backtest: bool
    debug: bool = False
    strategy_path: str



from enum import Enum
class SERVICES(Enum):
    historical_data_feeds='historical_data_feeds'
    live_data_feeds='live_data_feeds'
    python_backtester='python_backtester'
    python_engine='python_engine'
    python_executor='python_executor'

SERVICES_ARRAY = [i for i in dir(SERVICES) if not i.startswith('__')]


from typing import Union
import pandas as pd
from os.path import join
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import InstrumentFile
from os import getenv
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.data_sources.tradingview.downloader import Interval as TRADINGVIEW_INTERVALS
from backtesterrb30.libs.data_sources.tradingview.downloader import TradingviewDownloader
from backtesterrb30.libs.utils.timestamps import datetime_to_timestamp


class TradingView(DataSource):
    INTERVALS= TRADINGVIEW_INTERVALS
    NAME='tradingview'

    def __init__(self, logger=print):
        super().__init__(False, logger)
        trading_view_user=getenv("trading_view_user")
        trading_view_password=getenv("trading_view_password")
        if trading_view_user == None or trading_view_password == None:
            raise Exception(self.NAME + ' No authorization heys provided')
        self.client = TradingviewDownloader(trading_view_user, trading_view_password)
        self.df_to_clip = None

    #override
    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        symbol, exchange, fut_contract = self.get_symbol_params(data.symbol)
        self.df_to_clip: pd.DataFrame = await self.client.get_hist(symbol ,
                datetime_to_timestamp(data.backtest_date_start), 
                exchange,
                data.interval,
                fut_contract)
        return True

    #override
    async def _download_instrument_data(self,
                    instrument: str, interval: str, time_start: int, time_stop: Union[int, None]) -> pd.DataFrame:
        self._log('Downloading tradingview data', instrument)
        if time_stop == None:
            df_orig = df[df['timestamp'] >= time_start]
            df = df_orig.iloc[:, [0,1]]
            milis = self._get_interval_miliseconds(interval)
            if type(milis) != int or milis == 0:
                self._log('Warning, binance cannot get close price')
            else:
                df_close = df_orig.iloc[-1:, [0,4]] 
                df_close.columns = df.columns
            df = pd.concat([df, df_close])
            # print(df)
        else:
            df = self.__clip_df(time_start, time_stop, self.df_to_clip)
        return df


    def _get_interval_miliseconds(self, interval: str) -> Union[int,None]: 
        in_1_minute = "1"
        in_3_minute = "3"
        in_5_minute = "5"
        in_15_minute = "15"
        in_30_minute = "30"
        in_45_minute = "45"
        in_1_hour = "1H"
        in_2_hour = "2H"
        in_3_hour = "3H"
        in_4_hour = "4H"
        in_daily = "1D"
        in_weekly = "1W"
        in_monthly = "1M"

    def __clip_df(self, time_start: int, time_stop: int, df: pd.DataFrame):
        df = df[df['timestamp'] >= time_start]
        df = df[df['timestamp'] <= time_stop]
        df = df.iloc[:-1, [1,2]]
        return df

    @staticmethod
    def create_symbol_name(symbol: str, exchange: str, fut_contract: str = ''):
        if not symbol or symbol == '':
            raise Exception('bad symbol name')
        if not exchange or exchange == '':
            raise Exception('bad exchange name')
        return f"{symbol}_{exchange}_{fut_contract}" if \
                fut_contract and fut_contract != '' else f"{symbol}_{exchange}"

    @staticmethod
    def get_symbol_params(symbol: str):
        value = symbol.split('_')
        if len(value) <2:
            raise Exception('Bad symbol')
        if len(value) == 2: value.append(None)
        return tuple(value)





from typing import Union
from datetime import datetime
import pandas as pd
from os.path import join
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import InstrumentFile
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
# from backtesterrb30.libs.utils.historical_sources import COINGECKO_INTERVALS
from backtesterrb30.libs.utils.timestamps import datetime_to_timestamp, timestamp_to_datetime
from pycoingecko import CoinGeckoAPI
from backtesterrb30.libs.utils.singleton import singleton
import time
import asyncio
from enum import Enum

class COINGECKO_INTERVALS_2(Enum):
    day4='day4'

class CoingeckoDataSource(DataSource):
    INTERVALS= COINGECKO_INTERVALS_2
    NAME='coingecko'

    def __init__(self, logger=print):
        super().__init__(False, logger)
        self.client = CoinGeckoAPI()
        self.__last_request_time = 0

    #override
    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        try:
            candles = self.client.get_coin_ohlc_by_id(data.symbol, vs_currency='usd', days='365')
        except Exception as e:
            if e.args[0]['error'] == 'Could not find coin with the given id':
                raise e
        df = pd.DataFrame(candles)
        if datetime_to_timestamp(data.backtest_date_start) < df.iloc[0,0]:
            raise(Exception("Error. First avaliable date of " , data.symbol, "is" , timestamp_to_datetime(df.iloc[0,0])))
        if datetime_to_timestamp(data.backtest_date_stop) > df.iloc[-1,0]:
            raise Exception("Error. Last avaliable date of " , data.symbol, "is" , timestamp_to_datetime(df.iloc[0,0]))
        return True
    
    #override
    async def _download_instrument_data(self,
                    instrument: str, interval: str, time_start: int, time_stop: Union[int, None]) -> pd.DataFrame:
        self._log('Downloading coingecko data', instrument)
        if interval == COINGECKO_INTERVALS_2.day4.value:
            actual_time = time.time()
            if actual_time - self.__last_request_time < 0.025:
                self._log('waitin')
                await asyncio.sleep(0.025 - (actual_time - self.__last_request_time))
            candles = self.client.get_coin_ohlc_by_id(instrument, vs_currency='usd', days='365')
            df_orig = pd.DataFrame(candles)
            df = df_orig[df_orig[0] >= time_start]
            if time_stop == None:
                df = df_orig.iloc[:-1, [0,1]]
            else:
                df = df[df[0] <= time_stop]
                df = df.iloc[:-1, [0,1]]
        else:
            self._raise_error('No such interval')
        return df

    def get_current_price(self, data_symbol: DataSymbol):
        price =self.client.get_price(data_symbol.symbol,vs_currencies='USD')
        return price['polkadot']['usd']

    def _get_interval_miliseconds(self, interval: str) -> Union[int,None]: 
        if interval == COINGECKO_INTERVALS_2.day4.value: return 1000*60*60*24*4




from datetime import datetime
from json import load
from os import system, remove, walk
from os.path import join
import shutil
from typing import Union
import pandas as pd
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import InstrumentFile
from os import getenv, getcwd
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.utils.singleton import singleton
import importlib.resources
import json
from enum import Enum
from backtesterrb30.libs.utils.timestamps import timestamp_to_datetime
from shutil import rmtree

# from backtesterrb30.historical_data_feeds.modules.utils import validate_dataframe_timestamps

class DUKASCOPY_INTERVALS_2(str, Enum):
    tick: str='tick'
    minute: str='minute'
    minute5: str='minute5'
    minute15: str='minute15'
    minute30: str='minute30'
    hour: str='hour'
    hour4: str='hour4'
    day: str='day'
    month: str='month'

class DukascopyDataSource(DataSource):
    INTERVALS= DUKASCOPY_INTERVALS_2
    NAME='dukascopy'


    def __init__(self, logger=print):
        super().__init__(False, logger)

    def __get_ducascopy_interval(self, interval: str) -> str:
        if interval == 'tick': return 'tick'
        if interval == 'minute': return 'm1'
        if interval == 'minute5': return 'm5'
        if interval == 'minute15': return 'm15'
        if interval == 'minute30': return 'm30'
        if interval == 'hour': return 'h1'
        if interval == 'day': return 'd1'
        if interval == 'month': return 'mn1'

    def _get_interval_miliseconds(self, interval: str) -> Union[int,None]: 
        if interval == 'tick': return None
        if interval == 'minute': return 60*1000
        if interval == 'minute5': return 5*60*1000
        if interval == 'minute15': return 15*60*1000
        if interval == 'minute30': return 30*60*1000
        if interval == 'hour': return 60*60*1000
        if interval == 'day': return 24*60*60*1000
        if interval == 'month': return None


    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        # https://raw.githubusercontent.com/Leo4815162342/dukascopy-node/master/src/utils/instrument-meta-data/generated/raw-meta-data-2022-04-23.json
        # response = requests.get("http://api.open-notify.org/astros.json")
        from_datetime_timestamp = int(round(datetime.timestamp(data.backtest_date_start) * 1000))
        # f = open('temporary_ducascopy_list.json')
        # instrument_list = load(f)['instruments']
        with importlib.resources.open_text("backtesterrb30", "temporary_ducascopy_list.json") as file:
            instrument_list = json.load(file)['instruments']
        #validate if instrument exists:
        if data.symbol.upper() not in [v['historical_filename'] for k, v in instrument_list.items()]:
            self._log('Error. Instrument "'+data.symbol+'" does not exists on dukascopy.')
            return False

        #validate it timestamps perios is right:
        for k, v in instrument_list.items():
            if v["historical_filename"] == data.symbol.upper():
                first_timestamp = int(v["history_start_day"])
                if first_timestamp > from_datetime_timestamp:
                    print("Error. First avaliable date of " , data.symbol, "is" , datetime.fromtimestamp(first_timestamp/1000.0))
                    return False

        return True

    async def _download_instrument_data(self, 
                    instrument: str, interval: str, time_start: int, time_stop: Union[int, None]) -> pd.DataFrame:
        self._log('Downloading dukascopy data', instrument)
        """
        documentation: 
        https://github.com/Leo4815162342/dukascopy-node
        """

        duca_interval = self.__get_ducascopy_interval(interval)
        time_start_datetime = timestamp_to_datetime(time_start)
        if [time_start_datetime.hour,
            time_start_datetime.minute,
            time_start_datetime.second,
            time_start_datetime.microsecond] != [0,0,0,0]:
            raise Exception('Cannot dowload data with this time_start! Intervals must be in day graduality')
        time_stop_datetime = timestamp_to_datetime(time_stop)
        if [time_stop_datetime.hour,
            time_stop_datetime.minute,
            time_stop_datetime.second,
            time_stop_datetime.microsecond] != [0,0,0,0]:
            raise Exception('Cannot dowload data with this time_stop! Intervals must be in day graduality')
        from_param = datetime.fromtimestamp(time_start//1000.0).strftime("%Y-%m-%d")
        to_param = datetime.fromtimestamp(time_stop//1000.0).strftime("%Y-%m-%d")
        if from_param == to_param: 
            raise Exception('The same start and stop date')
        here = getcwd()
        cache_path = join(here, 'cache_dukascopy')
        rmtree(cache_path)
        # system('rm -r ' + cache_path)
        string_params = [
            ' -i '+ instrument,
            ' -from '+ from_param,
            ' -to '+ to_param,
            ' -s',
            ' -t ' + duca_interval,
            ' -fl', 
            ' -f csv',
            ' -dir '+ cache_path,
            ' -p bid'
        ]
        command = 'npx dukascopy-node'
        for param in string_params:
            command += param
        system(command)
        name_of_created_file = next(walk(cache_path), (None, None, []))[2][0]  
        df = pd.read_csv(join(cache_path, name_of_created_file), index_col=None, header=None)
        if duca_interval == 'tick': 
            df = df.iloc[1:, [0,2]]
        else:
            df = df.iloc[1:, [0,1]]
        if name_of_created_file and name_of_created_file != '':
            remove(join(cache_path, name_of_created_file))
        # print('Dukascopy df shape after dl', df.shape)
        # print('head', str(df.head(1)))
        # print('tail', str(df.tail(1)))
        return df
        


from pydantic import BaseModel

from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol

class Trade(BaseModel):
    timestamp: int
    value: float
    price: float
    symbol: str
    source: str
    # data_symbol: DataSymbol


from pydantic import BaseModel
from typing import Optional, List
from backtesterrb30.libs.interfaces.python_engine.custom_chart_element import CustomChartElement


class CustomChart(BaseModel):
    name: str
    display_on_price_chart: Optional[bool] = False
    log_scale: Optional[bool] = False
    color: Optional[str] = 'orange'
    chart: List[CustomChartElement]


from typing import Union
from xmlrpc.client import DateTime
from binance import Client
from binance.helpers import interval_to_milliseconds
from datetime import datetime
import pandas as pd
from os.path import join
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import InstrumentFile
# from backtesterrb30.libs.utils.historical_sources import BINANCE_INTERVALS
# from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.data_sources.utils import validate_dataframe_timestamps
from os import getenv
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.utils.singleton import singleton
from enum import Enum

class BINANCE_INTERVALS_2(str, Enum):
    tick: str='tick'
    minute: str='minute'
    minute3: str='minute3'
    minute5: str='minute5'
    minute15: str='minute15'
    minute30: str='minute30'
    hour: str='hour'
    hour2: str='hour2'
    hour4: str='hour4'
    hour6: str='hour6'
    hour8: str='hour8'
    hour12: str='hour12'
    day: str='day'
    day3: str='day3'
    week: str='week'
    month: str='month'

class BinanceDataSource(DataSource):
    INTERVALS= BINANCE_INTERVALS_2
    NAME='binance'

    def __init__(self, logger=print):
        super().__init__(False, logger)
        binance_api_secret=getenv("binance_api_secret")
        binance_api_key=getenv("binance_api_key")
        if binance_api_secret == None or binance_api_key == None:
            raise Exception(self.NAME + ' No authorization heys provided')
        self.client = Client(binance_api_key, binance_api_secret)
        self._test_counter = 0

    #override
    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        #validate if instrument exists:
        exchange_info = self.client.get_exchange_info()
        if data.symbol not in [s['symbol'] for s in exchange_info['symbols']]:
            self._log('Error. Instrument "'+data.symbol+'" does not exists on binance.')
            return False
        #validate it timestamps perios is right:
        from_datetime_timestamp = int(round(datetime.timestamp(data.backtest_date_start) * 1000))
        if data.interval == BINANCE_INTERVALS_2.tick:
            binance_interval = Client.KLINE_INTERVAL_1MINUTE
        else:
            binance_interval = self.__get_binance_interval(data.interval.value)
        first_timestamp = self.client._get_earliest_valid_timestamp(data.symbol, binance_interval)
        if first_timestamp > from_datetime_timestamp:
            self._log("Error. First avaliable date of " , data.symbol, "is" , datetime.utcfromtimestamp(first_timestamp/1000.0))
            return False
        return True
    
    #override
    async def _download_instrument_data(self,
                        instrument: str, interval: str, time_start: int, time_stop: Union[int, None]) -> pd.DataFrame:
        self._log('Downloading binance data', instrument)
        if interval == 'tick': 
            interval_timestamp = 1000*60*60
            trades_arr = []
            start_hour = time_start
            stop_hour = start_hour + interval_timestamp
            while stop_hour < time_stop:
                trades = self.client.get_aggregate_trades(symbol=instrument, 
                        startTime=start_hour, endTime=stop_hour)
                # print('trades', trades)
                trades_arr = trades_arr + trades
                start_hour += interval_timestamp
                stop_hour += interval_timestamp
            df = pd.DataFrame(trades_arr)
            df = df[['T','p']]
        else:
            binance_interval = self.__get_binance_interval(interval)
            klines = self.client.get_historical_klines(instrument, 
                    binance_interval, time_start, time_stop)
            df_orig = pd.DataFrame(klines)
            df = df_orig.iloc[:-1, [0,1]]
            if time_stop == None:
                df = df_orig.iloc[:, [0,1]]
                milis = self._get_interval_miliseconds(interval)
                if type(milis) != int or milis == 0:
                    self._log('Warning, binance cannot get close price')
                else:
                    df_close = df_orig.iloc[-1:, [0,4]] 
                    df_close.columns = df.columns
                df = pd.concat([df, df_close])

            # if interval != STRATEGY_INTERVALS.tick.value:
            #     df = validate_dataframe_timestamps(df, interval, time_start, time_stop)

            # print('binance shape after dl: ', df.shape)
            # print('head', str(df.head(1)))
            # print('tail', str(df.tail(1)))

        return df

    def get_current_price(self, symbol: DataSymbol):
        # self._test_counter += 1
        # print('binance get curren price', self._test_counter)
        price = self.client.get_symbol_ticker(symbol = symbol.symbol)
        return price['price']

    def _get_interval_miliseconds(self, interval: str) -> Union[int,None]: 
        return interval_to_milliseconds(self.__get_binance_interval(interval))

    def __get_binance_interval(self, interval: str) -> str :
        if interval == 'minute': return Client.KLINE_INTERVAL_1MINUTE
        if interval == 'minute3': return Client.KLINE_INTERVAL_3MINUTE
        if interval == 'minute5': return Client.KLINE_INTERVAL_5MINUTE
        if interval == 'minute15': return Client.KLINE_INTERVAL_15MINUTE
        if interval == 'minute30': return Client.KLINE_INTERVAL_30MINUTE
        if interval == 'hour': return Client.KLINE_INTERVAL_1HOUR
        if interval == 'hour2': return Client.KLINE_INTERVAL_2HOUR
        if interval == 'hour4': return Client.KLINE_INTERVAL_4HOUR
        if interval == 'hour6': return Client.KLINE_INTERVAL_6HOUR
        if interval == 'hour8': return Client.KLINE_INTERVAL_8HOUR
        if interval == 'hour12': return Client.KLINE_INTERVAL_12HOUR
        if interval == 'day': return Client.KLINE_INTERVAL_1DAY
        if interval == 'day3': return Client.KLINE_INTERVAL_3DAY
        if interval == 'week': return Client.KLINE_INTERVAL_1WEEK
        if interval == 'month': return Client.KLINE_INTERVAL_1MONTH




from abc import abstractmethod
import asyncio
from typing import List, Union
from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.interfaces.python_backtester.custom_chart import CustomChart
from backtesterrb30.libs.interfaces.python_backtester.data_finish import DataFinish 
from backtesterrb30.libs.interfaces.python_backtester.debug_breakpoint import DebugBreakpoint
from backtesterrb30.libs.interfaces.python_backtester.last_feed import LastFeed
from backtesterrb30.libs.interfaces.python_engine.custom_chart_element import CustomChartElement
from backtesterrb30.libs.interfaces.python_engine.price_event import PriceEvent
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.communication_broker.zmq_broker import ZMQ
from backtesterrb30.libs.communication_broker.asyncio_broker import AsyncioBroker
from backtesterrb30.libs.utils.list_of_services import SERVICES
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.utils.module_loaders import import_spec_module, reload_spec_module
from backtesterrb30.libs.utils.json_serializable import JSONSerializable
from backtesterrb30.libs.utils.service import Service
from backtesterrb30.libs.interfaces.utils.config import Config, BROKERS



class Engine(Service):
    """Python Engine"""
    _broker: BrokerBase
    
    def __init__(self, config: Config, data_schema: DataSchema, loop = None, logger=print):
        super().__init__(config, logger)
        self.config: Config=config
        self.__loop =  loop
        self.__custom_event_loop = False
        if self.__loop == None: 
            self.__loop = asyncio.get_event_loop()
            self.__custom_event_loop= True
        self.__data_schema: DataSchema = data_schema
        self.__columns=['timestamp']+[c.symbol for c in self.__data_schema.data]
        self.__data_buffer = [ [] for col in self.__columns]
        for sym, arr in zip(self.__data_schema.data, self.__data_buffer[1:]):
            sym.additional_properties['buffer'] = arr
            sym.additional_properties['price_events'] = []
        self.__buffer_length = 1
        self.__custom_charts: List[CustomChart] = []
        self.__debug_mode = False
        self.__debug_next_pressed = False
        self.__reloading_modules = []
        self.__send_breakpoint_after_feed = False
        self.__code_stopped_debug = False
        self.__breakpoint_display_charts = True

    # public methods:
    # ==================================================================

    @abstractmethod
    async def on_feed(self, data):
        """Function is being called when new data frame appears.
        While defining strategy you have to override this function
        and define your strategy logic here. 

        :param data: data buffer defined in :class:`DataSchema`
        :type data: list
        """
        pass

    def on_data_finish(self):
        """Function is being called in backtest mode when historical data are finished
        """
        pass


    def get_data_schema(self) -> DataSchema:
        """Returns data_schema defined in your strategy files in `data_schema.py`
        Data schema contains list of :class:`DataSymbol` objects. In every :class:`DataSymbol` object, 
        you can use `get_buffer()` function to get current buffer connected to this symbol.

        :return: :class:`DataSchema` object combined with this strategy.
        :rtype: DataSchema
        """
        return self.__data_schema


    def get_data_symbol_by_custom_name(self, custom_name: str) -> DataSymbol:
        """Returns a :class:`DataSymbol` object using `custom_name` defined
        in `data_schema.py`, `custom_name` must be unique.

        :param custom_name: custom name defined in `data_schema.py` file.
        :type custom_name: str
        :return: :class:`DataSymbol` object combined with privided custom_name
        :rtype: DataSymbol
        """
        if type(custom_name) != str:
            raise Exception('Provided name is not string')
        arr = [d for d in self.__data_schema.data if d.custom_name == custom_name]
        if len(arr) == 0:
            raise Exception('No data symbol with such custom name')
        if len(arr) >1 : 
            raise Exception('Two elements with the same custom name')
        return arr[0]


    def get_columns(self) -> List[str]:
        """Returns a list of instrument names based on all instruments in :class:`DataSchema`.
        Do not use it as identifiers! Names can be the same in other data sources.

        :return: List of returned :class:`str` objects
        :rtype: list
        """
        return self.__columns


    def set_buffer_length(self, length: int):
        """Sets data buffer length. Data provided to `on_feed()` function
        is going to have this length.

        :param length: buffer length
        :type length: int
        """
        if length < 1: 
            raise Exception('Buffer length must be bigget than 1')
        # if self.config.backtest == False and length > 1:
        #     raise Exception('Buffer longer than 1 not implemended in live mode')
        self.__buffer_length = length

 
    async def trigger_event(self, event: JSONSerializable):
        """Calling this function, you trigger `on_event()` function in 
        :class:`TradeExecutor`. You can provide any JSON serializable input
        to this method. 

        :param event: any JSON serializable data
        :type event: JSONSerializable
        """
        await self.__send_last_feed(SERVICES.python_executor)
        await self._broker.send(SERVICES.python_executor,'event', event)
    

    def add_custom_chart(self, 
                    chart: List[CustomChartElement], 
                    name: str, 
                    display_on_price_chart: Union[DataSymbol, None] = None, 
                    log_scale: Union[bool, None] = None, 
                    color: Union[str, None] = None):
        """Function adds custom chart printed under main charts in summary
        or in breakpoint debug event.

        :param chart: list of :class:`CustomChartElement`
        :type chart: list
        :param name: name of your custom name
        :type name: str
        :param display_on_price_chart: determining if chart is going to be displayed on other chart (on privided :class:`DataSymbol` chart) or it is going to be displayed alone
        :type display_on_price_chart: DataSymbol, optional
        :param log_scale: determines if chart is going to be in log scale
        :type log_scale: bool, optional
        :param color: a list of UUID strings, defaults to None
        :type color: str, optional
        """
        chart_obj = {
            'chart': chart,
            'display_on_price_chart': display_on_price_chart,
            'name': name
        }
        if display_on_price_chart: chart_obj['display_on_price_chart'] = display_on_price_chart
        if log_scale: chart_obj['log_scale'] = log_scale
        if color: chart_obj['color'] = color
        self.__custom_charts.append(chart_obj)

    
    async def debug_breakpoint(self, display_charts = True):
        """This function causes breakpoint triggered while you turn on debug mode.

        :param display_charts: determines if breakpoint should cause displaying charts
        :type display_charts: bool, optional
        """   
        self.__breakpoint_display_charts = display_charts
        if self.__debug_mode == True:
            self.__code_stopped_debug = True
            while True:
                if self.__debug_mode == False:
                    for spec, module in self.__reloading_modules:
                        # reload(module)
                        reload_spec_module(spec, module)
                        # spec.loader.exec_module(module)

                        # module = SourceFileLoader('strategies.hackaton1.reloading_module.reloading_methods', '/home/george/workspace/project/Retire-Before-30/Engine-RB30/strategies/hackaton1/reloading_module/reloading_methods.py')

                    self.__code_stopped_debug = False
                    return
                if self.__debug_next_pressed == True:
                    for spec, module in self.__reloading_modules:
                        # reload(module)
                        reload_spec_module(spec, module)
                        spec.loader.exec_module(module)
                        # module = SourceFileLoader('strategies.hackaton1.reloading_module.reloading_methods', '/home/george/workspace/project/Retire-Before-30/Engine-RB30/strategies/hackaton1/reloading_module/reloading_methods.py')

                    self.__debug_next_pressed = False
                    self.__send_breakpoint_after_feed = True
                    self.__code_stopped_debug = False
                    return 
                await asyncio.sleep(0.1)


    def add_reloading_module(self, module_path: str):
        """Adds module that is going to be reloaded 
        every step of your debug. Module must not save any state
        because while realoading this state is going to be refreshed.

        :param module_path: path to module
        :type module_path: str
        """
        spec, module = import_spec_module(module_path)
        self.__reloading_modules.append((spec, module))
        reload_spec_module(spec,module)
        return module

    # ==================================================================
    # end of public methods
    

    #private methods:

    def _loop(self):
        # self._broker.run()
        self._broker.create_listeners(self.__loop)
        if self.config.debug == True:
            self.__loop.create_task(self.__keyboard_listener())
        if self.__custom_event_loop:
            self.__loop.run_forever()
            self.__loop.close()

    # def _send(self, service: SERVICES, msg: str, *args):
    #     self._broker.send(service, msg, *args)

    def _configure(self):
        super()._configure()
        self._broker.register("data_feed", self.__data_feed_event)
        self._broker.register("historical_sending_locked", self.__historical_sending_locked_event)
        self._broker.register("data_finish", self.__data_finish_event)
        self._broker.register("engine_ready", self.__engine_ready_event)
        self._broker.register("get_buffer_length", self.__get_buffer_length_event)
        self._broker.register("register_price_event", self.__register_price_event)

    # def _send(): pass
    # def _register(): pass
    # def _create_listeners(): pass


    # # override
    # def _asyncio_loop(self, loop: asyncio.AbstractEventLoop):
    #     self._broker._create_listeners(loop)
    #     loop.create_task(self.__keyboard_listener())

    # # override
    # def _handle_zmq_message(self, message):
    #     pass


    async def __send_last_feed(self, service: SERVICES):
        last_feed = {
            'last_feed': [v[-1] for v in self.__data_buffer]
        }
        await self._broker.send(service, 'last_feed', LastFeed(**last_feed))


    async def __send_debug_breakpoint(self):
        breakpoint_params= {}
        breakpoint_params['custom_charts'] = self.__custom_charts
        breakpoint_params['display_charts'] = self.__breakpoint_display_charts 
        self._log('sending debug breakpoint')
        await self.__send_last_feed(SERVICES.python_executor)
        await self._broker.send(SERVICES.python_executor,'debug_breakpoint', DebugBreakpoint(**breakpoint_params))


    async def __keyboard_listener(self):
        import keyboard
        self._log('To enter debug mode press "ctrl+d"')
        while True:
            if keyboard.is_pressed('ctrl+d'): 
                if self.__debug_mode == False:
                    self._log('You have entered Debug mode \n\
                                 -> press "ctrl+n" to next step\n\
                                 -> press "ctrl+q" to leave debug mode')
                    self.__debug_mode = True
                while keyboard.is_pressed('d'):
                    await asyncio.sleep(0.1)
            if keyboard.is_pressed('ctrl+n'):
                if self.__debug_mode == True and self.__debug_next_pressed == False and self.__code_stopped_debug:
                    self._log('next step ... \n\
                                -> press "ctrl+q" to leave debug mode')
                    self.__debug_next_pressed = True
                while keyboard.is_pressed('n'):
                    await asyncio.sleep(0.1)
            if keyboard.is_pressed('ctrl+q'):
                if self.__debug_mode == True:
                    self._log('You have leaved Debug mode \n\
                                -> press "ctrl+d" to enter debug mode again')
                    self.__debug_mode = False
                while keyboard.is_pressed('q'):
                    await asyncio.sleep(0.1)
            await asyncio.sleep(0.1)


    #COMMANDS

    async def __data_feed_event(self, new_data_row):
        for i, v in enumerate(new_data_row):
            self.__data_buffer[i].append(v)
        if len(self.__data_buffer[0])>self.__buffer_length:
            for i, v in enumerate(new_data_row):
                self.__data_buffer[i].pop(0)
        if len(self.__data_buffer[0])>=self.__buffer_length:
            await self.on_feed(self.__data_buffer)
            if self.__send_breakpoint_after_feed: 
                await self.__send_debug_breakpoint()
                self.__send_breakpoint_after_feed = False

        
        
    async def __historical_sending_locked_event(self):
        await self._broker.send(SERVICES.historical_data_feeds,'unlock_historical_sending')

    
    async def __data_finish_event(self):
        self.on_data_finish()
        self.__debug_mode = False
        finish_params = {}
        finish_params['custom_charts'] = self.__custom_charts
        # finish_params['main_instrument_price'] = self.__get_main_intrument_price()
        await self.__send_last_feed(SERVICES.python_backtester)
        await self._broker.send(SERVICES.python_backtester, 'data_finish', DataFinish(**finish_params))

    async def __engine_ready_event(self, service_name):
        await self._broker.send(SERVICES[service_name], 'engine_ready_response')

        
    async def __get_buffer_length_event(self, service_name):
        await self._broker.send(SERVICES[service_name], 'engine_set_buffer_length', self.__buffer_length)


    async def __register_price_event(self, price_event):
        price_event = PriceEvent(**price_event)
        symbol: DataSymbol = [data for data in self.__data_schema.data if data.identifier == \
                price_event.symbol_itentifier][0]
        symbol.additional_properties['price_events'].append(price_event)


from datetime import timezone
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema


def validate_config(config: dict):
    for d in config['data']:
        if type(d['historical_data_source']) != str:
            d['historical_data_source'] = d['historical_data_source'].NAME
    cfg = DataSchema(**config)
    for symbol in cfg.data:
        symbol.backtest_date_start = symbol.backtest_date_start.replace(tzinfo=timezone.utc)
        symbol.backtest_date_stop = symbol.backtest_date_stop.replace(tzinfo=timezone.utc)
    return cfg